@@ -16,9 +16,14 @@ from django.db import connection, router, models
 from django.db.models.deletion import DO_NOTHING
 from django.db.models.fields.files import FieldFile, FileDescriptor
 from django.db.models.fields.related import RelatedField, Field, ManyToOneRel
-from django.contrib.contenttypes.generic import GenericRelation, GenericRel
+from django.contrib.contenttypes.generic import GenericRelation
 from django.utils.functional import curry
 
+try:
+    from django.contrib.contenttypes.fields import GenericRel
+except ImportError:
+    from django.contrib.contenttypes.generic import GenericRel
+
 try:
     from django.db.models.fields.related import ForeignObject, ForeignObjectRel
 except ImportError:
@@ -387,7 +392,7 @@ class GenericForeignFileDescriptor(object):
 
         qn = connection.ops.quote_name
 
-        if hasattr(instance._default_manager, 'prefetch_related'):
+        if hasattr(instance.__class__._default_manager, 'prefetch_related'):
             # Django 1.4+
             manager_kwargs = {
                 'prefetch_cache_name': self.field.attname,
@@ -478,7 +483,20 @@ class GenericForeignFileDescriptor(object):
         # handle None.
         attr_cls = self.file_field.attr_class
 
-        if isinstance(value, self.field.rel.to):
+        # Because of the (some would say boneheaded) way pickle works,
+        # the underlying FieldFile might not actually itself have an associated
+        # file. So we need to reset the details of the FieldFile in those cases.
+        if isinstance(value, attr_cls):
+            value.instance = instance
+            value.field = self.file_field
+            value.storage = self.file_field.storage
+            value.related_object = obj
+            instance.__dict__[self.file_field.name] = value
+            return
+
+        if isinstance(obj, self.field.rel.to):
+            value = getattr(obj, self.field.rel_file_field_name)
+        elif isinstance(value, self.field.rel.to):
             obj = value
             value = getattr(obj, self.field.rel_file_field_name)
 
@@ -488,26 +506,22 @@ class GenericForeignFileDescriptor(object):
             instance.__dict__[self.file_field.name] = attr
 
         # Other types of files may be assigned as well, but they need to have
-        # the FieldFile interface added to the. Thus, we wrap any other type of
-        # File inside a FieldFile (well, the field's attr_class, which is
-        # usually FieldFile).
-        elif isinstance(value, File) and not isinstance(value, FieldFile):
+        # the correct FieldFile interface added to them. Thus, we wrap any
+        # other type of File inside a FieldFile (well, the field's attr_class,
+        #  which is usually FieldFile).
+        elif isinstance(value, File):
             file_copy = attr_cls(instance, self.file_field, value.name)
-            file_copy.file = value
+            if isinstance(value, FieldFile):
+                # Avoid unnecessary IO caused by accessing ``value.file``
+                if value and getattr(value, '_file', None):
+                    file_copy.file = value.file
+                file_copy._committed = value._committed
+            else:
+                file_copy.file = value
+                file_copy._committed = False
             file_copy.related_object = obj
-            file_copy._committed = False
             instance.__dict__[self.file_field.name] = file_copy
 
-        # Finally, because of the (some would say boneheaded) way pickle works,
-        # the underlying FieldFile might not actually itself have an associated
-        # file. So we need to reset the details of the FieldFile in those cases.
-        elif isinstance(value, FieldFile):
-            value.instance = instance
-            value.field = self.file_field
-            value.storage = self.file_field.storage
-            value.related_object = obj
-            instance.__dict__[self.file_field.name] = value
-
     def __set__(self, instance, value):
         if instance is None:
             raise AttributeError("Manager must be accessed via instance")
